@@ -211,8 +211,7 @@ read_token(const int sock,
 	 * Check and make sure token looks right
 	 */
 	if (((header[flag] < 20) || (header[flag] > 26)) ||
-	    (header[major_version] != 3) ||
-	    ((header[minor_version] != 0) && (header[minor_version] != 1))) {
+	    (header[major_version] != 3)) {
 	    if (*p_buffer != NULL) {
             free(*p_buffer);
             *p_buffer = NULL;
@@ -986,44 +985,57 @@ GSI_SOCKET_get_peer_name(GSI_SOCKET *self,
 char *
 GSI_SOCKET_get_peer_hostname(GSI_SOCKET *self)
 {
-    struct sockaddr_in		addr;
+    struct sockaddr_storage addr;
     socklen_t			addr_len = sizeof(addr);
-    struct hostent		*info;
+    char host			[NI_MAXHOST];
+    int					loopback=0;
 
     if (getpeername(self->sock, (struct sockaddr *) &addr,
 		    &addr_len) < 0) {
-	self->error_number = errno;
-	GSI_SOCKET_set_error_string(self, "Could not get peer address");
-	return NULL;
+        self->error_number = errno;
+        GSI_SOCKET_set_error_string(self, "Could not get peer address");
+        return NULL;
     }
 
-    info = gethostbyaddr((char *)&addr.sin_addr,
-			 sizeof(addr.sin_addr),
-			 addr.sin_family);
-    if ((info == NULL) || (info->h_name == NULL)) {
-	self->error_number = errno;
-	GSI_SOCKET_set_error_string(self, "Could not get peer hostname");
-	return NULL;
+    if (getnameinfo((struct sockaddr *) &addr, addr_len,
+                    host, sizeof(host),
+                    NULL, 0, NI_NAMEREQD)) {
+        self->error_number = errno;
+        GSI_SOCKET_set_error_string(self, "Could not get peer hostname");
+        return NULL;
     }
 
-    if (info->h_addrtype == AF_INET) { /* check for localhost */
-	struct in_addr inaddr;
-	inaddr = *(struct in_addr *)(info->h_addr);
-	if (ntohl(inaddr.s_addr) == INADDR_LOOPBACK) {
-	    char buf[MAXHOSTNAMELEN];
-	    if (gethostname(buf, sizeof(buf)) < 0) {
-		self->error_number = errno;
-		GSI_SOCKET_set_error_string(self, "gethostname() failed");
-		return NULL;
-	    }
-	    info = gethostbyname(buf);
-	    if (info == NULL || info->h_name == NULL) {
-		return strdup(buf);
-	    }
-	}
+    /* check for localhost / loopback */
+    if (addr.ss_family == AF_INET) {
+        struct sockaddr_in sadder;
+
+        memcpy(&sadder, &addr, sizeof(sadder));
+        if (ntohl(sadder.sin_addr.s_addr) == INADDR_LOOPBACK) {
+            loopback = 1;
+        }
+    }
+#ifdef AF_INET6
+    else if (addr.ss_family == AF_INET6) {
+        struct sockaddr_in6 saddr6;
+
+        memcpy(&saddr6, &addr, sizeof(saddr6));
+        if (IN6_IS_ADDR_LOOPBACK(&saddr6.sin6_addr)) {
+            loopback = 1;
+        }
+    }
+#endif
+
+    if (loopback) {
+        char buf[MAXHOSTNAMELEN];
+        if (gethostname(buf, sizeof(buf)) < 0) {
+            self->error_number = errno;
+            GSI_SOCKET_set_error_string(self, "gethostname() failed");
+            return NULL;
+        }
+        return strdup(buf);
     }
 
-    return strdup(info->h_name);
+    return strdup(host);
 }
 
 #ifdef HAVE_VOMS
@@ -1033,9 +1045,14 @@ add_fqan(char ***fqans, const char *fqan)
    int current_len;
    char **new_fqans;
 
+   if (fqans==NULL) {
+      return GSI_SOCKET_ERROR;
+   }
+
+
    current_len = 0;
-   if (fqans != NULL) {
-      while (fqans[current_len] != NULL)
+   if (*fqans != NULL) {
+      while ((*fqans)[current_len] != NULL)
 	 current_len++;
    }
 
@@ -1050,6 +1067,58 @@ add_fqan(char ***fqans, const char *fqan)
 
    return 0;
 }
+
+static gss_OID_desc gss_ext_x509_cert_chain_oid_desc =
+     {11, "\x2b\x06\x01\x04\x01\x9b\x50\x01\x01\x01\x08"}; 
+static gss_OID_desc * gss_ext_x509_cert_chain_oid =
+                &gss_ext_x509_cert_chain_oid_desc;
+
+static int
+GSI_SOCKET_get_peer_cert_chain(GSI_SOCKET *self,
+                               X509 **cert,
+                               STACK_OF(X509) **cert_chain)
+{
+    OM_uint32 major_status = 0;
+    OM_uint32 minor_status = 0;
+    gss_buffer_set_t buffer_set = NULL;
+    int i;
+
+    *cert = NULL;
+    *cert_chain = NULL;
+
+    major_status = gss_inquire_sec_context_by_oid(&minor_status,
+                                                  self->gss_context,
+                                                  gss_ext_x509_cert_chain_oid,
+                                                  &buffer_set);
+    if (major_status != GSS_S_COMPLETE) {
+        GSI_SOCKET_set_error_string(self, "gsi_inquire_sec_context_by_oid() failed in GSI_SOCKET_get_peer_cert_chain()");
+        return GSI_SOCKET_ERROR;
+    }
+
+    *cert_chain = sk_X509_new_null();
+
+    for (i = 0; i < buffer_set->count; i++) {
+        const unsigned char *p;
+        X509 *c;
+                    
+        p = buffer_set->elements[i].value;
+        c = d2i_X509(NULL, &p, buffer_set->elements[i].length);
+
+        if (i == 0) {
+            *cert = c;
+        } else {
+            if (sk_X509_insert(*cert_chain,
+                               c, sk_X509_num(*cert_chain)) == SSL_ERROR) {
+                GSI_SOCKET_set_error_string(self, "sk_X509_insert() failed in GSI_SOCKET_get_peer_cert_chain()");
+                gss_release_buffer_set(&minor_status, &buffer_set);
+                return GSI_SOCKET_ERROR;
+            }
+        }
+    }
+
+    gss_release_buffer_set(&minor_status, &buffer_set);
+    return GSI_SOCKET_SUCCESS;
+}
 #endif
 
 int
@@ -1066,6 +1135,8 @@ GSI_SOCKET_get_peer_fqans(GSI_SOCKET *self, char ***fqans)
    char **fqan = NULL;
    int voms_err;
    char *err_msg, *err_str;
+   X509 *cert = NULL;
+   STACK_OF(X509) *cert_chain = NULL;
 
    voms_data = VOMS_Init(NULL, NULL);
    if (voms_data == NULL) {
@@ -1074,8 +1145,16 @@ GSI_SOCKET_get_peer_fqans(GSI_SOCKET *self, char ***fqans)
       return GSI_SOCKET_ERROR;
    }
 
-   ret = VOMS_RetrieveFromCtx(self->gss_context, RECURSE_CHAIN,
-	                      voms_data, &voms_err);
+   if (GSI_SOCKET_get_peer_cert_chain(self,
+                                      &cert,
+                                      &cert_chain) != GSI_SOCKET_SUCCESS) {
+      GSI_SOCKET_set_error_string(self, "Failed to read VOMS attributes, GSI_SOCKET_get_peer_cert_chain( failed");
+      return GSI_SOCKET_ERROR;
+   }
+
+   ret = VOMS_Retrieve(cert,
+                       cert_chain,
+                       RECURSE_CHAIN, voms_data, &voms_err);
    if (ret == 0) {
       if (voms_err == VERR_NOEXT) {
 	 /* No VOMS extensions present, return silently */
@@ -1106,6 +1185,10 @@ GSI_SOCKET_get_peer_fqans(GSI_SOCKET *self, char ***fqans)
 end:
    if (voms_data)
       VOMS_Destroy(voms_data);
+   if (cert)
+       X509_free(cert);
+   if (cert_chain)
+       sk_X509_pop_free(cert_chain, X509_free);
 
    return ret;
 #endif
