@@ -175,7 +175,7 @@ int
 main(int argc, char *argv[]) 
 {    
     pid_t childpid, otherpid;
-    struct sockaddr_in client_addr;
+    struct sockaddr_storage client_addr;
     socklen_t client_addr_len = sizeof(client_addr);
     sigset_t mysigset;
     struct pidfh *pfh = NULL;
@@ -272,7 +272,14 @@ main(int argc, char *argv[])
     }
 
     if (!server_context->run_as_daemon) {
-       myproxy_log("Connection from %s", inet_ntoa(client_addr.sin_addr));
+       server_context->usage.client_ip[0] = '\0';
+       getnameinfo((struct sockaddr *)&client_addr,
+                   sizeof(client_addr),
+                   server_context->usage.client_ip,
+                   sizeof(server_context->usage.client_ip),
+                   NULL, 0,
+                   NI_NUMERICHOST);
+       myproxy_log("Connection from %s", server_context->usage.client_ip);
        socket_attrs->socket_fd = fileno(stdin);
        if (handle_client(socket_attrs, server_context) < 0) {
 	  my_failure("error in handle_client()");
@@ -371,9 +378,13 @@ main(int argc, char *argv[])
 	     }
 	     
 	     /* child process */
-             snprintf(server_context->usage.client_ip,
-                      sizeof(server_context->usage.client_ip), "%s",
-                      inet_ntoa(client_addr.sin_addr)); /* TODO: IPv6 support */
+         server_context->usage.client_ip[0] = '\0';
+         getnameinfo((struct sockaddr *)&client_addr,
+                     sizeof(client_addr),
+                     server_context->usage.client_ip,
+                     sizeof(server_context->usage.client_ip),
+                     NULL, 0,
+                     NI_NUMERICHOST);
              myproxy_log("Connection from %s", server_context->usage.client_ip);
 	     close(0);
 	     close(1);
@@ -398,6 +409,9 @@ main(int argc, char *argv[])
 
  parent_exit:
     pidfile_remove(pfh);
+#ifdef HAVE_GLOBUS_USAGE
+    myproxy_usage_stats_close(server_context);
+#endif
     return 0;
 }   
 
@@ -1005,6 +1019,81 @@ init_arguments(int argc, char *argv[],
     return arg_error;
 }
 
+static int
+bind_socket(const char *hostname, int port)
+{
+    int sock=-1;
+    struct addrinfo hints, *res, *ressave;
+    int on = 1;
+    struct linger lin = {0,0};
+    int n;
+    char portstr[6]={0}, *portstrp=NULL;
+
+    /* getaddrinfo() requires either hostname or port to be set */
+    assert(hostname || port); 
+        
+    if (port) {
+        snprintf(portstr, 6, "%d", port);
+        portstrp=portstr;
+    }
+
+    memset(&hints, 0, sizeof(struct addrinfo));
+    hints.ai_flags    = AI_PASSIVE;
+    hints.ai_family   = AF_UNSPEC;
+    hints.ai_socktype = SOCK_STREAM;
+
+    n = getaddrinfo(hostname, portstrp, &hints, &res);
+    if (n < 0) {
+        myproxy_log("getaddrinfo error: %s", gai_strerror(n));
+        return -1;
+    }
+    ressave=res;
+
+    while (res) {
+        char chosenhost[NI_MAXHOST] = { 0 };
+        char chosenport[NI_MAXSERV] = { 0 };
+
+        getnameinfo(res->ai_addr, res->ai_addrlen,
+                    chosenhost, sizeof(chosenhost),
+                    chosenport, sizeof(chosenport),
+                    NI_NUMERICHOST|NI_NUMERICSERV);
+
+        sock = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+
+        if (sock >= 0) {
+
+            /* Allow reuse of socket */
+            setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
+            setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *)&lin, sizeof(lin));
+
+            if (bind(sock, res->ai_addr, res->ai_addrlen) == 0) {
+                myproxy_log("Socket bound to %s:%s", chosenhost, chosenport);
+                break;
+            }
+
+            if (errno == EADDRINUSE) {
+                myproxy_log("Port %s on %s already in use, probably by another "
+                    "myproxy-server instance.\nUse the -p option to run "
+                    "multiple myproxy-server instances on different "
+                    "ports.", chosenport, chosenhost);
+            } else {
+                myproxy_log("Failed to bind socket to %s:%s: %s",
+                            chosenhost, chosenport, strerror(errno));
+            }
+            close(sock);
+            sock=-1;
+        } else {
+            myproxy_log("Failed to create socket for %s:%s: %s",
+                        chosenhost, chosenport, strerror(errno));
+        }
+        res = res->ai_next;
+    }
+
+    freeaddrinfo(ressave);
+
+    return sock;
+}
+
 /*
  * myproxy_init_server()
  *
@@ -1015,13 +1104,8 @@ init_arguments(int argc, char *argv[],
 int 
 myproxy_init_server(myproxy_socket_attrs_t *attrs) 
 {
-    int on = 1;
-    int listen_sock;
-    struct sockaddr_in sin;
-    socklen_t socklen;
-    struct linger lin = {0,0};
+    int listen_sock=-1;
     GSI_SOCKET *tmp_gsi_sock;
-    struct hostent *hp;
 
     if ((tmp_gsi_sock = GSI_SOCKET_new(0)) == NULL) {
 	failure("malloc() failed in GSI_SOCKET_new()");
@@ -1036,62 +1120,54 @@ myproxy_init_server(myproxy_socket_attrs_t *attrs)
     }
     GSI_SOCKET_destroy(tmp_gsi_sock);
     
-    listen_sock = socket(AF_INET, SOCK_STREAM, 0);
+    if (attrs->pshost || attrs->psport) {
+        myproxy_debug("using getaddrinfo() to configure listen socket");
+        listen_sock = bind_socket(attrs->pshost, attrs->psport);
+    } else { /* just create unbound IPv4 socket for now */
+        int on = 1;
+        struct linger lin = {0,0};
 
+        myproxy_debug("creating IPv4 listen socket without binding");
+        listen_sock = socket(AF_INET, SOCK_STREAM, 0);
+
+        /* Allow reuse of socket */
+        setsockopt(listen_sock,
+                   SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
+        setsockopt(listen_sock,
+                   SOL_SOCKET, SO_LINGER, (char *)&lin, sizeof(lin));
+    }
     if (listen_sock == -1) {
-        failure("Error in socket()");
+        failure("Error creating server socket");
     } 
 
-    /* Allow reuse of socket */
-    setsockopt(listen_sock, SOL_SOCKET, SO_REUSEADDR, (void *) &on, sizeof(on));
-    setsockopt(listen_sock, SOL_SOCKET, SO_LINGER, (char *) &lin, sizeof(lin));
-
-    memset(&sin, 0, sizeof(sin));
-    sin.sin_family = AF_INET;
-    sin.sin_port = htons(attrs->psport);
-    if (attrs->pshost == NULL) {
-        sin.sin_addr.s_addr = INADDR_ANY;
-    } else {
-        /* First, try inet_addr to see if pshost is an IP addr */
-        sin.sin_addr.s_addr = inet_addr(attrs->pshost);
-        if (sin.sin_addr.s_addr == -1) {
-            /* pshost was not valid IP addr, so try gethostbyname */
-            hp = gethostbyname(attrs->pshost);
-            if (hp != NULL) {
-                /* Got resolv of hostname, so use it */
-                memcpy(&(sin.sin_addr.s_addr),hp->h_addr,hp->h_length);
-            } else {
-                failure("Hostname specified by --listen is invalid.");
-            }
-        }
-    }
-
-    if (bind(listen_sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
-	if (errno == EADDRINUSE) {
-	    myproxy_log("Port %d on %s already in use, probably by another "
-			"myproxy-server instance.\nUse the -p option to run "
-			"multiple myproxy-server instances on different "
-			"ports.", attrs->psport, 
-            ((attrs->pshost == NULL) ? "localhost" : attrs->pshost) );
-	}
-	failure("Error in bind()");
-    }
     if (listen(listen_sock, INT_MAX) < 0) {
 	    failure("Error in listen()");
     }
 
     if (attrs->psport == 0) {
-        memset(&sin, 0, sizeof(sin));
-        socklen = sizeof(sin);
-        if (getsockname(listen_sock, (struct sockaddr *) &sin, &socklen) < 0) {
+        struct sockaddr_storage addr;
+        socklen_t socklen = sizeof(addr);
+
+        if (getsockname(listen_sock, (struct sockaddr *) &addr, &socklen) < 0) {
             failure("Error in getsockname()");
         }
-        attrs->psport = ntohs(sin.sin_port);
+        if (addr.ss_family == AF_INET) {
+            struct sockaddr_in saddr;
+            memcpy(&saddr, &addr, sizeof(saddr));
+            attrs->psport = ntohs(saddr.sin_port);
+        }
+#ifdef AF_INET6
+        else if (addr.ss_family == AF_INET6) {
+            struct sockaddr_in6 saddr6;
+            memcpy(&saddr6, &addr, sizeof(saddr6));
+            attrs->psport = ntohs(saddr6.sin6_port);
+        }
+#endif
     }
 
     /* Got this far? Then log success! */
     myproxy_log("Starting myproxy-server on %s:%d...",
-                ((attrs->pshost == NULL) ? "localhost" : attrs->pshost),
+                ((attrs->pshost == NULL) ? "*" : attrs->pshost),
                 attrs->psport);
 
     return listen_sock;
